from .base import Detector
from .mediapipe import MediapipeDetector


detectors = {
    "mediapipe": MediapipeDetector
}


def get_detectors_names() -> list[str]:
    return list(detectors.keys())


def get_detector(name: str) -> Detector | None:
    if name in detectors:
        return detectors[name]()
    return None



from time import sleep
import os
from contextlib import redirect_stdout, redirect_stderr

from PIL import Image
import mediapipe as mp
import numpy as np

from .base import Detector


class MediapipeDetector(Detector):
    
    def __init__(self):
        self.hands = mp.solutions.hands.Hands(
            static_image_mode=True,
            max_num_hands=2,
            min_detection_confidence=0.1
        )
        self.face_detection = mp.solutions.face_detection.FaceDetection(
            min_detection_confidence=1.0
        )
    
    def detect(self, image: Image) -> bool:
        sleep(0.5)
        with open(os.devnull, 'w') as fnull:
            with redirect_stdout(fnull), redirect_stderr(fnull):
                fingertips_positions = self._get_fingertips_positions(image)
                mouth_position, mouth_size = self._get_mouth_position_and_size(image)
                if mouth_position is None:
                    return False
                closest_distance = self._get_closest_distance(fingertips_positions, mouth_position)
                return closest_distance < mouth_size / 4

    def _get_fingertips_positions(self, image: Image):
        image_rgb = np.asarray(image)
        hands_result = self.hands.process(image_rgb)
        
        all_positions = []
        if hands_result.multi_hand_landmarks:
            for hand_landmarks in hands_result.multi_hand_landmarks:
                positions = [hand_landmarks.landmark[i] for i in [4, 8, 12, 16, 20]]
                positions = [(int(landmark.x*image.width), int(landmark.y*image.height)) for landmark in positions]
                all_positions.extend(positions)
                    
        return all_positions
    
    def _get_mouth_position_and_size(self, image: Image):
        image_rgb = np.asarray(image)
        face_result = self.face_detection.process(image_rgb)
        
        max_area = 0
        position = None
        size = None
        
        if face_result.detections:
            for detection in face_result.detections:
                bbox = detection.location_data.relative_bounding_box
                area = bbox.width * bbox.height
                if area > max_area:
                    max_area = area
                    pos = detection.location_data.relative_keypoints[3]
                    position = (int(pos.x*image.width), int(pos.y*image.height))
                    size = int(bbox.width * image.width)
                    
        return position, size

    def _get_closest_distance(self, points: list, point: tuple) -> float:
        min_dist = float('inf')
        for p in points:
            dist = np.linalg.norm(np.array(p) - np.array(point))
            min_dist = min(min_dist, dist)
        return min_dist



from pathlib import Path
from time import sleep

import pygame

from .base import Alert


class BeepAlert(Alert):
    
    def __init__(self):
        super().__init__()
        pygame.mixer.init()
        sound_path = str(Path(__file__).parent.parent / "assets" / "error.wav")
        self.sound = pygame.mixer.Sound(sound_path)
    
    def on_start(self):
        self.sound.play()
        sleep(0.5)
    
    def on_active(self):
        self.sound.play()
        sleep(0.5)



import click

from nailguard.detectors import get_detectors_names, get_detector
from nailguard.alerts import get_alerts_names, get_alert
from nailguard.nailguard import Nailguard


@click.command()
@click.option(
    "--detector",
    type=click.Choice(get_detectors_names()),
    multiple=True,
    default=["mediapipe"],
    help="Detector to use. Defaults to mediapipe."
)
@click.option(
    "--alert",
    type=click.Choice(get_alerts_names()),
    multiple=True,
    default=["beep"],
    help="Alerts to use. Defaults to notification, beep."
)
@click.option(
    "--camera",
    type=int,
    default=0,
    help="Camera index to use. Defaults to 0."
)
@click.option(
    "--trigger",
    type=int,
    default=1,
    help="Number of detectors which must trigger to alert. Defaults to 1."
)
@click.option(
    "--debounce",
    type=float,
    default=1.0,
    help="Number of second to wait before firing alert. Defaults to 1.0"
)
def main(detector: str, alert: list[str], camera: int, trigger: int, debounce: float):
    """Nailguard launcher"""
    
    detectors = [get_detector(detector) for detector in detector]
    alerts = [get_alert(alert) for alert in alert]
    Nailguard(detectors, alerts, camera, trigger, debounce).run()



from pathlib import Path
from time import sleep

from notifypy import Notify

from .base import Alert


class NotificationAlert(Alert):
    
    def __init__(self):
        super().__init__()
        self.icon_path = str(Path(__file__).parent.parent / "assets" / "stop.png")
    
    def on_start(self):
        notification = Notify()
        notification.title = "Nailguard"
        notification.message = "Don't bite your nails!"
        notification.application_name = "Nailguard"
        notification.icon = self.icon_path
        notification.send()



class Alert:
    
    def __init__(self):
        self.last_detected = False
    
    def handle(self, detected: bool):
        if detected and not self.last_detected:
            self.on_start()
        elif not detected and self.last_detected:
            self.on_stop()
        elif detected:
            self.on_active()
        self.last_detected = detected
    
    def on_start(self):
        pass

    def on_stop(self):
        pass
    
    def on_active(self):
        pass



from .base import Alert
from .notification import NotificationAlert
from .beep import BeepAlert

alerts = {
    "notification": NotificationAlert,
    "beep": BeepAlert,
}


def get_alerts_names() -> list[str]:
    return list(alerts.keys())


def get_alert(name: str) -> Alert | None:
    if name in alerts:
        return alerts[name]()
    return None



import threading
from time import sleep

import cv2
from PIL import Image
from nailguard.detectors import Detector
from nailguard.alerts import Alert


IMAGE_READ_FPS = 5
ALERT_CHECK_FPS = 10
MASTER_DETECTOR_FPS = 10


class Nailguard:
    
    def __init__(
        self,
        detectors: list[Detector],
        alerts: list[Alert],
        camera_idx: int,
        trigger_count: int,
        debounce: float
    ) -> None:
        self.detectors = detectors
        self.alerts = alerts
        self.camera_idx = camera_idx
        self.trigger_count = trigger_count
        self.debounce = debounce
        
        self.image = None
        self.current_detected = {detector: False for detector in detectors}
        self.master_detected = False
    
    def run(self) -> None:
        threading.Thread(target=self._image_thread, args=(self,)).start()
        
        for detector in self.detectors:
            threading.Thread(target=self._detector_thread, args=(self, detector)).start()
            
        threading.Thread(target=self._master_detector_thread, args=(self,)).start()
        
        for alert in self.alerts:
            threading.Thread(target=self._alert_thread, args=(self, alert)).start()

    @staticmethod
    def _image_thread(nailguard) -> None:
        cap = cv2.VideoCapture(nailguard.camera_idx)
        while True:
            ret, image = cap.read()
            if not ret:
                raise Exception("Failed to capture image from camera")
            
            image = Image.fromarray(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))
            nailguard.image = image
            sleep(1 / IMAGE_READ_FPS)
        
    @staticmethod
    def _detector_thread(nailguard, detector: Detector) -> None:
        while True:
            if nailguard.image is None:
                continue
            detected = detector.detect(nailguard.image)
            nailguard.current_detected[detector] = detected

    @staticmethod
    def _master_detector_thread(nailguard) -> None:
        master_detector_period = 1 / MASTER_DETECTOR_FPS
        detected_count = 0
        while True:
            detected = all(nailguard.current_detected.values())
            if detected:
                detected_count += 1
            else:
                detected_count = 0
            nailguard.master_detected = detected_count > nailguard.debounce / master_detector_period
            sleep(master_detector_period)

    @staticmethod
    def _alert_thread(nailguard, alert: Alert) -> None:
        sleep_time = 1 / ALERT_CHECK_FPS
        while True:
            alert.handle(nailguard.master_detected)
            if not nailguard.master_detected:
                sleep(sleep_time)
